use crate::{lexer::LexerError, span::Span};

/// Kinds of errors that can occur during parsing.
#[derive(Debug, Clone, PartialEq)]
pub enum ParserErrorKind {
    /// An error which ocurred during lexical analysis.
    Lexer(LexerError),
    /// Expected a value for key in map, but no value was found.
    MissingValueInMap,
    /// An unexpecting closing parenthesis/brace/bracket was encountered.
    UnexpectedCloseBracket,
    /// Unexpectedly reached the end of the input.
    UnexpectedEof,
    /// Unexpected parser state reached.
    UnexpectedState,
    /// Opening parenthesis/brace/bracket does not have a matching closing
    /// parenthesis/brace/bracket.
    UnclosedBracket,
    /// An unmatched parenthesis/brace/bracket was encountered.
    UnmatchedBracket,
}

/// Parser error, with a start and end location.
#[derive(Debug, Clone, PartialEq)]
pub struct ParserError {
    /// The type of error encountered.
    pub kind: ParserErrorKind,
    /// The span in which the error occurred.
    pub span: Span,
}

impl ParserError {
    /// Construct a new instance of `ParserErorr`.
    #[must_use]
    pub const fn new(kind: ParserErrorKind, span: Span) -> Self {
        Self { kind, span }
    }
}

impl std::error::Error for ParserError {}

#[cfg(not(tarpaulin_include))]
impl std::fmt::Display for ParserError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        use ParserErrorKind::*;

        match &self.kind {
            Lexer(err) => write!(f, "{err}"),
            MissingValueInMap => write!(f, "Key in map is missing its value"),
            UnexpectedCloseBracket => write!(f, "Unexpected closing bracket"),
            UnexpectedEof => write!(f, "Unexpected reached end of input"),
            UnexpectedState => write!(f, "Unexpected parsing state reached"),
            UnclosedBracket => write!(f, "Unclosed bracket"),
            UnmatchedBracket => write!(f, "Unmatched bracket"),
        }
    }
}

impl From<LexerError> for ParserError {
    fn from(err: LexerError) -> Self {
        let span = err.span.clone();
        Self::new(ParserErrorKind::Lexer(err), span)
    }
}
