pub use self::{
    error::{ParserError, ParserErrorKind},
    node::{Node, NodeKind},
};
use crate::lexer::{Lexer, TokenKind};

mod error;
mod node;

/// Parser for converting Onihime source code into an Abstract Syntax Tree
/// (AST).
#[derive(Debug)]
pub struct Parser<'parser> {
    lexer: Lexer<'parser>,
}

impl<'parser> Parser<'parser> {
    /// Create a new parser instance from a string.
    #[must_use]
    pub fn new(input: &'parser str) -> Self {
        Self {
            lexer: Lexer::new(input),
        }
    }

    /// Set the name of the lexer's source.
    pub fn set_name(&mut self, name: String) {
        self.lexer.set_name(name);
    }

    /// Parse the input string into an AST.
    pub fn parse(&mut self) -> Result<Vec<Node>, ParserError> {
        let mut parents = Vec::new();
        let mut cur_node = Node::new(NodeKind::List(Vec::new()), self.lexer.span());

        while let Some(token) = self.lexer.read()? {
            match token.kind {
                TokenKind::BlockComment(_) | TokenKind::LineComment(_) => {}

                TokenKind::OpenParen => {
                    let child = Node::new(NodeKind::List(Vec::new()), token.span);
                    parents.push(cur_node);
                    cur_node = child;
                }
                TokenKind::CloseParen => {
                    let mut parent = parents.pop().ok_or_else(|| {
                        ParserError::new(
                            ParserErrorKind::UnexpectedCloseBracket,
                            token.span.clone(),
                        )
                    })?;

                    cur_node.span.extend(&token.span);

                    if !matches!(cur_node.kind, NodeKind::List(_)) {
                        return Err(ParserError::new(
                            ParserErrorKind::UnmatchedBracket,
                            token.span,
                        ));
                    }

                    parent.push_node(cur_node)?;
                    cur_node = parent;
                }

                TokenKind::OpenBrace => {
                    let child = Node::new(NodeKind::Set(Vec::new()), token.span);
                    parents.push(cur_node);
                    cur_node = child;
                }
                TokenKind::OpenHashBrace => {
                    let child = Node::new(NodeKind::Map(Vec::new()), token.span);
                    parents.push(cur_node);
                    cur_node = child;
                }
                TokenKind::CloseBrace => {
                    let mut parent = parents.pop().ok_or_else(|| {
                        ParserError::new(
                            ParserErrorKind::UnexpectedCloseBracket,
                            token.span.clone(),
                        )
                    })?;

                    cur_node.span.extend(&token.span);

                    if !matches!(cur_node.kind, NodeKind::Map(_) | NodeKind::Set(_)) {
                        return Err(ParserError::new(
                            ParserErrorKind::UnmatchedBracket,
                            token.span,
                        ));
                    }

                    if let NodeKind::Map(ref vec) = cur_node.kind {
                        if vec.len() % 2 != 0 {
                            return Err(ParserError::new(
                                ParserErrorKind::MissingValueInMap,
                                token.span,
                            ));
                        }
                    }

                    parent.push_node(cur_node)?;
                    cur_node = parent;
                }

                TokenKind::OpenBracket => {
                    let child = Node::new(NodeKind::Vector(Vec::new()), token.span);
                    parents.push(cur_node);
                    cur_node = child;
                }
                TokenKind::CloseBracket => {
                    let mut parent = parents.pop().ok_or_else(|| {
                        ParserError::new(
                            ParserErrorKind::UnexpectedCloseBracket,
                            token.span.clone(),
                        )
                    })?;

                    cur_node.span.extend(&token.span);

                    if !matches!(cur_node.kind, NodeKind::Vector(_)) {
                        return Err(ParserError::new(
                            ParserErrorKind::UnmatchedBracket,
                            token.span,
                        ));
                    }

                    parent.push_node(cur_node)?;
                    cur_node = parent;
                }

                _ => {
                    let node = Node::try_from(token)?;
                    cur_node.push_node(node)?;
                }
            }
        }

        if !parents.is_empty() {
            return Err(ParserError::new(
                ParserErrorKind::UnclosedBracket,
                cur_node.span,
            ));
        }

        if let NodeKind::List(body) = cur_node.kind {
            Ok(body)
        } else {
            unreachable!() // In theory, at least...
        }
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    use crate::{
        lexer::{LexerError, LexerErrorKind, Symbol},
        parser::node::Atom,
        span::Span,
    };

    macro_rules! test {
        ( $name:ident: $input:literal, $src:ident => $ast:expr ) => {
            #[test]
            fn $name() {
                let mut parser = Parser::new($input);
                let $src = parser.lexer.source();
                assert_eq!(parser.parse(), $ast);
            }
        };
    }

    test!(empty: "", _src => Ok(vec![]));

    test!(error_invalid_number: "(+ 1.2.3)", src => Err(ParserError::new(
        ParserErrorKind::Lexer(LexerError::new(
            LexerErrorKind::InvalidNumber("1.2.3".into()),
            Span::new(3..8, src.clone())
        )),
        Span::new(3..8, src)
    )));

    test!(list: "(+ 1 2) ; sneaky comment :)", src => Ok(vec![
        Node::new(
            NodeKind::List(vec![
                Node::new(Atom::Symbol(Symbol::from("+")).into(), Span::new(1..2, src.clone())),
                Node::new(Atom::Integer(1).into(), Span::new(3..4, src.clone())),
                Node::new(Atom::Integer(2).into(), Span::new(5..6, src.clone())),
            ]),
            Span::new(0..7, src)
        )
    ]));

    test!(error_list_unmatched_bracket: "(]", src => Err(ParserError::new(
        ParserErrorKind::UnmatchedBracket,
        Span::new(1..2, src),
    )));

    test!(error_list_missing_close_paren: "(true", src => Err(ParserError::new(
        ParserErrorKind::UnclosedBracket,
        Span::new(0..1, src),
    )));

    test!(error_list_unexpected_close_paren: ")", src => Err(ParserError::new(
        ParserErrorKind::UnexpectedCloseBracket,
        Span::new(0..1, src)
    )));

    test!(map: "#{:a 0.0 :b 1.0}", src => Ok(vec![
        Node::new(
            NodeKind::Map(vec![
                Node::new(Atom::Keyword(Symbol::from("a")).into(), Span::new(2..4, src.clone())),
                Node::new(Atom::Float(0.0).into(), Span::new(5..8, src.clone())),
                Node::new(Atom::Keyword(Symbol::from("b")).into(), Span::new(9..11, src.clone())),
                Node::new(Atom::Float(1.0).into(), Span::new(12..15, src.clone())),
            ]),
            Span::new(0..16, src)
        )
    ]));

    test!(error_map_missing_value: "#{:x}", src => Err(ParserError::new(
        ParserErrorKind::MissingValueInMap,
        Span::new(4..5, src.clone())
    )));

    test!(error_map_unmatched_bracket: "#{)", src => Err(ParserError::new(
        ParserErrorKind::UnmatchedBracket,
        Span::new(2..3, src),
    )));

    test!(error_map_missing_close_brace: "#{", src => Err(ParserError::new(
        ParserErrorKind::UnclosedBracket,
        Span::new(0..2, src),
    )));

    test!(set: "{{} nil}", src => Ok(vec![
        Node::new(
            NodeKind::Set(vec![
                Node::new(NodeKind::Set(vec![]), Span::new(1..3, src.clone())),
                Node::new(NodeKind::Atom(Atom::Nil), Span::new(4..7, src.clone())),
            ]),
            Span::new(0..8, src),
        )
    ]));

    test!(error_set_unmatched_bracket: "{]", src => Err(ParserError::new(
        ParserErrorKind::UnmatchedBracket,
        Span::new(1..2, src),
    )));

    test!(error_set_missing_close_brace: "{", src => Err(ParserError::new(
        ParserErrorKind::UnclosedBracket,
        Span::new(0..1, src),
    )));

    test!(error_map_set_unexpected_close_brace: "}", src => Err(ParserError::new(
        ParserErrorKind::UnexpectedCloseBracket,
        Span::new(0..1, src)
    )));

    test!(vector: "['a' 'b' 'c']", src => Ok(vec![
        Node::new(
            NodeKind::Vector(vec![
                Node::new(NodeKind::Atom(Atom::Char('a')), Span::new(1..4, src.clone())),
                Node::new(NodeKind::Atom(Atom::Char('b')), Span::new(5..8, src.clone())),
                Node::new(NodeKind::Atom(Atom::Char('c')), Span::new(9..12, src.clone())),
            ]),
            Span::new(0..13, src),
        )
    ]));

    test!(error_vector_unmatched_bracket: "[}", src => Err(ParserError::new(
        ParserErrorKind::UnmatchedBracket,
        Span::new(1..2, src),
    )));

    test!(error_vector_missing_close_bracket: "{", src => Err(ParserError::new(
        ParserErrorKind::UnclosedBracket,
        Span::new(0..1, src),
    )));

    test!(error_vector_unexpected_close_bracket: "]", src => Err(ParserError::new(
        ParserErrorKind::UnexpectedCloseBracket,
        Span::new(0..1, src)
    )));

    test!(multiple_expressions: "(/ 6 3 (+ 1 2)) (* 2 5)\n(- 10 5)", src => Ok(vec![
        Node::new(
            NodeKind::List(vec![
                Node::new(Atom::Symbol(Symbol::from("/")).into(), Span::new(1..2, src.clone())),
                Node::new(Atom::Integer(6).into(), Span::new(3..4, src.clone())),
                Node::new(Atom::Integer(3).into(), Span::new(5..6, src.clone())),
                Node::new(
                    NodeKind::List(vec![
                        Node::new(Atom::Symbol(Symbol::from("+")).into(), Span::new(8..9, src.clone())),
                        Node::new(Atom::Integer(1).into(), Span::new(10..11, src.clone())),
                        Node::new(Atom::Integer(2).into(), Span::new(12..13, src.clone())),
                    ]),
                    Span::new(7..14, src.clone())
                ),
            ]),
            Span::new(0..15, src.clone())
        ),
        Node::new(
            NodeKind::List(vec![
                Node::new(Atom::Symbol(Symbol::from("*")).into(), Span::new(17..18, src.clone())),
                Node::new(Atom::Integer(2).into(), Span::new(19..20, src.clone())),
                Node::new(Atom::Integer(5).into(), Span::new(21..22, src.clone())),
            ]),
            Span::new(16..23, src.clone())
        ),
        Node::new(
            NodeKind::List(vec![
                Node::new(Atom::Symbol(Symbol::from("-")).into(), Span::new(25..26, src.clone())),
                Node::new(Atom::Integer(10).into(), Span::new(27..29, src.clone())),
                Node::new(Atom::Integer(5).into(), Span::new(30..31, src.clone())),
            ]),
            Span::new(24..32, src)
        ),
    ]));

    test!(function_application: "(join \"foo\" \"bar\")", src => Ok(vec![Node::new(
        NodeKind::List(vec![
            Node::new(Atom::Symbol(Symbol::from("join")).into(), Span::new(1..5, src.clone())),
            Node::new(Atom::String("foo".into()).into(), Span::new(6..11, src.clone())),
            Node::new(Atom::String("bar".into()).into(), Span::new(12..17, src.clone())),
        ]),
        Span::new(0..18, src)
    )]));
}
